/*
    remote-sensor, a wireless sensor data collection system
    Copyright (C) 2008  christian mock <cm@tahina.priv.at>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/


#include <avr/io.h>
//#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/power.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>

#include "hardware.h"
#include "rf12.h"
#include "sht.h"

/* EEPROM addresses */
uint8_t EEMEM device_id_ee = 1;
uint8_t EEMEM test_magic_ee = 0; // 0xc3;

/* globals */

uint8_t testmode;


struct { // see ../../README
  uint8_t sender;
  uint8_t seq;
  struct {
    uint8_t len:2;
    uint8_t type:6;
  } reading_type_v;
  uint16_t voltage;
  struct {
    uint8_t len:2;
    uint8_t type:6;
  } reading_type_sht;
  uint16_t sht_temp;
  uint16_t sht_humid;
} data = { 0, 0, { 1, 0 }, 0, {3, 0}, 0, 0 };



/* do ADC conversion of Vbg vs Vcc */
uint16_t sample_vcc(void) {
  uint16_t res;

  power_adc_enable();
  set_sleep_mode(SLEEP_MODE_ADC);
  ADCSRA |= _BV(ADEN) | _BV(ADSC);
  while(ADCSRA & _BV(ADSC))
    sleep_mode();
  res = ADC;
  ADCSRA &= ~(_BV(ADEN));
  power_adc_disable();
  return res;
}

/* actually, nominal 64 seconds */
void wait_a_minute() {
  uint8_t i;
  for(i = 8; i; i--) {
    wdt_enable(WDTO_8S);
    WDTCR |= _BV(WDIE);
    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    sleep_mode();
  }
  wdt_disable();
}

void wait_a_second() {
  wdt_enable(WDTO_1S);
  WDTCR |= _BV(WDIE);
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_mode();
}

int main(void) {
  wdt_disable();

  DDRB = _BV(PB1) | _BV(PB2) | _BV(PB3) | _BV(PB4);
  // MOSI, SCK/shtDATA, /SEL, shtSCK
  PORTB = _BV(PB3); // deselect RFM12

  /* ADC: measure 1.1V internal reference against Vcc */
  /* prescaler /8 -> 125khz */
  /* single conv mode */
  ADMUX = _BV(MUX3) | _BV(MUX2);
  ADCSRA = _BV(ADIF) | _BV(ADIE) | /* clear interrupt flag, enable interrupt */
    _BV(ADPS1) | _BV(ADPS0);	/* prescaler /8 = 125kHz */
  ADCSRB = 0;

  power_all_disable();
  power_usi_enable();

  data.sender = eeprom_read_byte(&device_id_ee);
  testmode = eeprom_read_byte(&test_magic_ee);

  rf12_init();
  power_usi_disable();
  sei();

  sht_init();
  
  while(1) {
    // sleep first, work later -- when batteries get empty, sensor may run
    // into a reset cycle and shouldn't continually xmit...
    if(testmode == 0xc3) 
      wait_a_second();
    else
      wait_a_minute();

    data.sht_temp = sht_measure(SHT_MEAS_TEMP);
    data.sht_humid = sht_measure(SHT_MEAS_HUMID);
    data.voltage = sample_vcc();

    power_usi_enable();
    DDRB |= _BV(PB2); // SCK as output after SHT code messed with it
    rf12_cmd(RF12_POWER_SETTING, RF12_ET | RF12_DC); // turn on xmit

    rf12_send((uint8_t *)&data, sizeof(data));
    // wait till the next-to-last byte is sent (the last is the dummy)
    while (!rf12_read_status_MSB()); 
    rf12_cmd(RF12_POWER_SETTING, RF12_DC); // turn off
    power_usi_disable();

    data.seq++;
  }
}

ISR(ADC_vect) { /* empty, but must exist for ADC conversion */
}

ISR(WDT_vect) {
}
