/*
    remote-sensor, a wireless sensor data collection system
    Copyright (C) 2008  christian mock <cm@tahina.priv.at>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/


/* daemon for the remote sensors
   cm@tahina.priv.at 2008
*/

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <math.h>
#include <unistd.h>
#include <syslog.h>
#include <usb.h> /* written for libusb 0.1.12 */
#include "sensord.h"

char usage[] = "usage: %s [-d cricketdir] [-c csvfile] [-f]\n -f: stay in foreground\n";

struct usb_bus *busses;

uint8_t last_seq[256];	       // last seen sequence number per sender
unsigned int lost[256];		// lost packets per sender
unsigned int dups[256];		// duplicate packets per sender
unsigned int msgs[256];		// number of messages per sender (not
				// including duplicates)

char *cricketdirname, *csvfilename;
int foreground;
FILE *csvf;

#define DEBUG if(foreground)

extern char *optarg;

usb_dev_handle *find_receiver(void) {
  struct usb_bus *bus;
  char Manufacturer[STRINGSZ], Product[STRINGSZ];
  struct usb_device *dev;
  struct usb_dev_handle *devh = NULL;

  usb_init();
  usb_find_busses();
  usb_find_devices();
  busses = usb_get_busses();

  for(bus = busses; bus; bus = bus->next) {
    for(dev = bus->devices; dev; dev = dev->next) {
      if(dev->descriptor.idVendor == USB_VENDOR &&
	 dev->descriptor.idProduct == USB_PRODUCT &&
	 dev->descriptor.bcdDevice == USB_PRODUCT_VERSION) {
	devh = usb_open(dev);
	if(!devh) {
	  syslog(LOG_ERR,"usb_open: %s", usb_strerror());
	  goto xit;
	}
	if(usb_get_string_simple(devh, dev->descriptor.iManufacturer,
				 Manufacturer, STRINGSZ-1) < 0) {
	  syslog(LOG_ERR,"usb_get_string_simple(iManufacturer): %s",
		 usb_strerror());
	  goto xit;
	}
	if(usb_get_string_simple(devh, dev->descriptor.iProduct, Product,
				 STRINGSZ-1) < 0) {
	  syslog(LOG_ERR,"usb_get_string_simple(iProduct): %s",
		 usb_strerror());
	  goto xit;
	}
	syslog(LOG_INFO, "Found device \"%s\" \"%s\"\n",
	       Manufacturer, Product);
	if(!strcmp(Manufacturer, USB_MANUF_STRING) &&
	   !strcmp(Product, USB_PRODUCT_STRING)) {
	  if(usb_set_configuration(devh, 1) < 0) {
	    syslog(LOG_ERR,"usb_set_configuration: %s", usb_strerror());
	    exit(3);
	  }
	  if(usb_claim_interface(devh, 0) < 0) {
	    syslog(LOG_ERR,"usb_claim_interface: %s", usb_strerror());
	    exit(3);
	  }
	  if((dev->config->interface->altsetting->endpoint->bmAttributes &
	      USB_ENDPOINT_TYPE_MASK) != USB_ENDPOINT_TYPE_INTERRUPT) {
	    syslog(LOG_ERR, "Endpoint 1 is not an interrupt endpoint");
	    goto xit;
	  }
	  return devh;
	}
      }
    xit:
      if(devh) {
	usb_release_interface(devh, 0);
	usb_close(devh);
	devh = NULL;
      }
    }
  }
  return NULL;
}

void close_receiver(usb_dev_handle *devh) {
  if(devh) {
    usb_release_interface(devh, 0);
    usb_close(devh);
  }
}

double raw2temp(uint16_t v) {
  return -39.6 + 0.01 * v;
}

double raw2humid(uint16_t v) {
  return -4.0 + 0.0405 * v + -2.8e-6 * v * v;
}

void convert_sht11_raw(float *temp, float *humid, float *dewp, int raw_temp,
		       int raw_humid) {  
  float H;

  *temp = -39.60 + 0.01 * raw_temp; // we assume Vcc = 3V here, see data sheet
  *humid = -4 + 0.0405 * raw_humid + -2.8e-6 * raw_humid * raw_humid;
  *humid = *humid + (*temp - 25) * (0.01 + 0.00008 * raw_humid);

  H = (log10f(*humid)-2)/0.4343 + (17.62 * *temp)/(243.12+ *temp);
  *dewp = 243.12*H/(17.62-H);
}

void handle_message(struct message *message, int msglen) {
  uint8_t l, *p;
  struct sensor_reading *r;
  int tmpfd, len, i;
  char tmpfn[STRINGSZ];
  char cricketfn[STRINGSZ];
  char msgdump[255*3+1]; // maxlen x "%02x "
  char timestamp[20]; // "yyyy-mm-dd hh:mm:ss\0"
  char *mp;
  FILE *tmpf = NULL;
  time_t now;
  mode_t cur_umask;
  float sht_temp, sht_humid, sht_dewp;

  len = msglen;
  
  // BUG: when we lose exactely 256 packets from a sender...
  if(last_seq[message->sender] == message->seq) { // duplicate
    dups[message->sender]++;
    return;
  }
  msgs[message->sender]++;
  l = message->seq - last_seq[message->sender] - 1;
  if(l && (message->seq > 5)) { // sender could have rebooted
    lost[message->sender] += l;
  }
  last_seq[message->sender] = message->seq;

  p = message->data; msglen -= 2;

  if(cricketdirname) {
    strcpy(tmpfn, cricketdirname);
    strcat(tmpfn, "/tmpXXXXXX");
    if(-1 == (tmpfd = mkstemp(tmpfn))) {
      syslog(LOG_ERR, "error from mkstemp: %m");
      exit(2);
    }
    if(!(tmpf = fdopen(tmpfd, "w"))) {
      syslog(LOG_ERR, "error from fdopen: %m");
      exit(2);
    }
    // mkstemp creates mode 0600; we overwrite with umask values...
    cur_umask = umask(022); umask(cur_umask);
    if(-1 == fchmod(tmpfd, 0666 & ~cur_umask)) {
      syslog(LOG_ERR, "error on chmod: %m");
      exit(2);
    }
    
    fprintf(tmpf, "%d # messages\n%d # duplicates\n%d # lost\n",
	    msgs[message->sender], dups[message->sender],
	    lost[message->sender]);
  }

  if(csvf) {
    now = time(NULL);
    strftime(timestamp, 20, "%F %T", localtime(&now));
    fprintf(csvf, "%s,%d,%d,%d,%d,%d,", timestamp, message->sender,
	    message->seq, msgs[message->sender], dups[message->sender],
	    lost[message->sender]);
  }

  while(msglen > 0) {
    r = (struct sensor_reading *)p;

    switch(r->len) {

    case 0:
      switch(r->type) {

      case 0:
	if(tmpf)
	  fprintf(tmpf, "%d # %% power out\n", r->data.one);
	if(csvf)
	  fprintf(csvf, "%d,", r->data.one);
	break;

      default:
	syslog(LOG_ERR, "unknown type:len %d:%d from %d",
	       r->type, r->len, message->sender);
      }
      break;

    case 1:
      switch(r->type) {

      case 0: // VCC_BG_11
	if(tmpf)
	  fprintf(tmpf, "%f # Vcc (BG 1.10)\n", 1.10 * 1023 / r->data.two);
	if(csvf)
	  fprintf(csvf, "%f,", 1.10 * 1023 / r->data.two);
	break;

      case 1: // VCC_BG_13
	if(tmpf)
	  fprintf(tmpf, "%f # Vcc (BG 1.30)\n", 1.30 * 1023 / r->data.two);
	if(csvf)
	  fprintf(csvf, "%f,", 1.30 * 1023 / r->data.two);
	break;

      case 2: // TEMP_100
	if(tmpf)
	  fprintf(tmpf, "%f # C\n", (int16_t)(r->data.two)/100.0);
	if(csvf)
	  fprintf(csvf, "%f,", (int16_t)(r->data.two)/100.0);
	break;

      default:
	syslog(LOG_ERR, "unknown type:len %d:%d from %d",
	       r->type, r->len, message->sender);
      }
      break;

    case 2:
      syslog(LOG_ERR, "unknown type:len %d:%d from %d",
	     r->type, r->len, message->sender);
      break;

    case 3:
      switch(r->type) {

      case 0: // SHT_RAW
	convert_sht11_raw(&sht_temp, &sht_humid, &sht_dewp,
			  r->data.two, *(&(r->data.two) + 1));
	if(tmpf)
	  fprintf(tmpf, "%f # C\n%f # %% rel. humidity\n%f # C dewpoint\n",
		  sht_temp, sht_humid, sht_dewp);
	if(csvf)
	  fprintf(csvf, "%f,%f,%f,", sht_temp, sht_humid, sht_dewp);
	break;

      default:
	syslog(LOG_ERR, "unknown type:len %d:%d from %d",
	       r->type, r->len, message->sender);
	break;
      }
    }
    p += (r->len + 2);
    msglen -= (r->len + 2);
  }
  if(msglen < 0) {
    syslog(LOG_ERR, "broken message from %d, seq %d, len %d, remaining %d",
	   message->sender, message->seq, len, msglen);
    mp = msgdump;
    for(i=0; i<len; i++) {
      sprintf(mp, "%02x ", ((unsigned char *)message)[i]);
      mp += 3;
    }
    syslog(LOG_DEBUG, "dump: %s", msgdump);
  }

  if(csvf) {
    fprintf(csvf,"\n");
    fflush(csvf);
  }

  if(tmpf) {
    sprintf(cricketfn, "%s/sensor%03d", cricketdirname, message->sender);
    if(-1 == rename(tmpfn, cricketfn)) {
      syslog(LOG_ERR, "rename(%s,%s): %m", tmpfn, cricketfn);
      exit(2);
    }
    fclose(tmpf);
  }
}
  

int main(int argc, char **argv) {
  usb_dev_handle *devh;
  int len;
  char buffer[MAX_PKT_SIZE], opt;

  while((opt = getopt(argc, argv, "c:d:f")) != -1) {
    switch(opt) {
    case 'd':
      cricketdirname = optarg;
      break;
    case 'c':
      csvfilename = optarg;
      break;
    case 'f':
      foreground = 1;
      break;
    default:
      fprintf(stderr, usage, argv[0]);
      exit(1);
    }
  }

  openlog("sensord", foreground ? LOG_PERROR|LOG_PID : LOG_PID,
	  LOG_DAEMON);

  if(!cricketdirname && !csvfilename && !foreground) {
    fprintf(stderr, usage, argv[0]);
    syslog(LOG_ERR, "at least one option must be given");
    exit(1);
  }

  if(cricketdirname) {
    // dir + "/tmpXXXXXX\0" for mkstemp(3)
    if((strlen(cricketdirname) + 11) > STRINGSZ) { 
      syslog(LOG_ERR, "cricket directory name too long");
      exit(1);
    }
    if(-1 == mkdir(cricketdirname, 0777)) {
      if(errno != EEXIST) {
	syslog(LOG_ERR, "error opening cricket directory %s: %m",
	       cricketdirname);
	exit(2);
      }
    }
    if(-1 == access(cricketdirname, R_OK | W_OK)) {
      syslog(LOG_ERR, "error in permissions for cricket directory %s: %m",
	     cricketdirname);
      exit(2);
    }
  }

  if(csvfilename) {
    if(!(csvf = fopen(csvfilename, "a"))) {
      syslog(LOG_ERR, "error opening csvfile %s: %m", csvfilename);
      exit(2);
    }
  }

  if(!foreground) {
    if(-1 == daemon(0,0)) {
      syslog(LOG_ERR, "daemon: %m");
      exit(2);
    }
  }

  if(!(devh = find_receiver())) {
    syslog(LOG_ERR, "No receiver found at startup, exiting.");
    exit(3);
  }

  while(1) {
    // poll the data...
    while(len = usb_control_msg(devh, USB_TYPE_VENDOR | USB_RECIP_DEVICE |
				USB_ENDPOINT_IN, REQ_READ, 0, 0, buffer,
				MAX_PKT_SIZE, 1000)) {
      if(len < 0) {
	syslog(LOG_ERR,"usb_control_msg: %s", usb_strerror());
	close_receiver(devh);
	devh = NULL;
	break;
      } else
	handle_message((struct message *)buffer, len);
    }

    // ...then wait for an interrupt from the receiver
    if(devh && ((len = usb_interrupt_read(devh, 1, buffer, 8, 60000)) < 0)) {
      if(errno != EAGAIN) {
	syslog(LOG_ERR, "usb_interrupt_read: %s", usb_strerror());
	close_receiver(devh);
	devh = NULL;
      }
    }
    if(!devh) { // need to reopen
      syslog(LOG_INFO, "trying to re-open USB connection...");
      while(!devh) {
	sleep(USB_REOPEN_TIMEOUT);
	devh = find_receiver();
      }
    }
  }
  close_receiver(devh);
}
    
